! Atom relaxation algorithm by (quasi) Newton-Raphson methods
! Numerical Recipe F77

        module atom_relax_function
          double precision, allocatable :: p(:),g(:)
          double precision, allocatable :: dg(:),hdg(:),hessin(:,:),xi(:)
        end module atom_relax_function

        subroutine atom_relax
        use variables
        use constants
        use atom_relax_function
        implicit double precision (a-h,o-z)
        integer :: iter,nmax,itmax
        double precision :: fret,gtol,func,EPS,STPMX,TOLX
        parameter (NMAX=50,ITMAX=200,STPMX=100.,EPS=3.e-8,TOLX=4.*EPS)
        integer i,its,j
        logical check
        double precision :: fac,fad,fae,fp,sum,sumdg,sumxi
        double precision :: alph
 
        alph = relax_factor * 1.0d-3

        nx=n*3
        if (.not. allocated (p)) then
           allocate (p(nx),g(nx),dg(nx),hdg(nx),hessin(nx,nx),xi(nx))
           hessin=0.0d0
           do i=1,nx
              hessin(i,i)=1.0d0 * alph
           enddo
           k=1
           do i=1,n
              p(k)=rx(i); p(k+1)=ry(i); p(k+2)=rz(i)
              g(k)=fx(i); g(k+1)=fy(i); g(k+2)=fz(i)
              k=k+3
           enddo
           go to 1000
        else
           dg=g
           k=1
           do i=1,n
              p(k)=rx(i); p(k+1)=ry(i); p(k+2)=rz(i)
              g(k)=fx(i); g(k+1)=fy(i); g(k+2)=fz(i)
              k=k+3
           enddo
        endif

        do i=1,nx
           dg(i)=g(i)-dg(i)
        enddo
        do i=1,nx
           hdg(i)=0.0d0
           do j=1,nx
              hdg(i)=hdg(i)+hessin(i,j)*dg(j)
           enddo
        enddo
        fac=0.0d0
        fae=0.0d0
        sumdg=0.0d0
        sumxi=0.0d0
        do i=1,nx
           fac=fac+dg(i)*xi(i)
           fae=fae+dg(i)*hdg(i)
           sumdg=sumdg+dg(i)**2
           sumxi=sumxi+xi(i)**2
        enddo
        if(fac.gt.sqrt(EPS*sumdg*sumxi))then
           fac=1.0d0/fac
           fad=1.0d0/fae
           do i=1,nx
              dg(i)=fac*xi(i)-fad*hdg(i)
           enddo
           do i=1,nx
              do j=i,nx
                 hessin(i,j)=hessin(i,j)+fac*xi(i)*xi(j) &
                      -fad*hdg(i)*hdg(j) &
                      +fae*dg(i)*dg(j)
                 hessin(j,i)=hessin(i,j)
              enddo
           enddo
        endif

1000    continue

        do i=1,nx
           xi(i)=0.0d0
           do j=1,nx
              xi(i)=xi(i)+hessin(i,j)*g(j)
           enddo
        enddo

        k=1
        do i=1,n
           rx(i)=rx(i)+xi(k)
           ry(i)=ry(i)+xi(k+1)
           rz(i)=rz(i)+xi(k+2)
           vx(i)=xi(k)/dt
           vy(i)=xi(k+1)/dt
           vz(i)=xi(k+2)/dt
           k=k+3
        enddo

        end
